class SurvivalModelExplainer:
    def __init__(
        self,
        model,
        data,
        y,
        predict_survival_function=None,
        predict_cumulative_hazard_function=None,
    ):
        self.model = model
        self.data = data
        self.y = y
        self.predict_survival_function = predict_survival_function
        self.predict_cumulative_hazard_function = predict_cumulative_hazard_function

    def predict(self, data, function_type):
        if function_type == "sf":
            if self.predict_survival_function is not None:
                return self.predict_survival_function(self.model, data)
            elif hasattr(self.model, "predict_survival_function"):
                return self.model.predict_survival_function(data)
            else:
                raise ValueError(
                    "Pass a predict survival function to `SurvivalModelExplainer`,\
                    e.g. `predict_survival_function=lambda m, d: m.predict_survival_function(d)`."
                )
        elif function_type == "chf":
            if self.predict_cumulative_hazard_function is not None:
                return self.predict_cumulative_hazard_function(self.model, data)
            elif hasattr(self.model, "predict_cumulative_hazard_function"):
                return self.model.predict_cumulative_hazard_function(data)
            else:
                raise ValueError(
                    "Pass a predict cumulative hazard function to `SurvivalModelExplainer`,\
                    e.g. `predict_cumulative_hazard_function=lambda m, d: m.predict_cumulative_hazard_function(d)`."
                )
        else:
            raise ValueError("function type needs to be one of `sf` or `chf`")



from copy import deepcopy
import itertools
import numpy as np
import pandas as pd
from tqdm import tqdm
import math
from scipy.integrate import trapezoid
from sklearn.metrics import r2_score
from sksurv.ensemble import RandomSurvivalForest
import shap
import warnings


def shap_tree_explainer(explainer, new_observation, function_type, aggregation_method, timestamps, **kwargs):
    if timestamps is not None:
        warnings.warn(
            "timestamps are ignored for calculation_method = 'treeshap' \n SurvSHAP(t) values are calculated for explainer.model.unique_times_"
        )
    timestamps = explainer.model.unique_times_

    if not isinstance(explainer.model, RandomSurvivalForest):
        raise TypeError("explained model must be of class sksurv.ensemble.RandomSurvivalForest")

    if function_type == "sf":
        start_index = 1
    elif function_type == "chf":
        start_index = 0

    all_function_vals = [f.y for f in explainer.predict(explainer.data, function_type)]
    baseline_fun = np.mean(all_function_vals, axis=0)
    target_fun = explainer.predict(new_observation, function_type)[0].y

    n_estimators = len(explainer.model.estimators_)
    tree_ensemble_model = {"trees": [estimator.tree_ for estimator in explainer.model.estimators_]}

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        exp = shap.TreeExplainer(tree_ensemble_model, explainer.data, **kwargs)
        res = exp.shap_values(new_observation)

    shap_values = np.vstack(res[start_index::2]).T / n_estimators

    variable_names = explainer.data.columns
    result = prepare_result_df(new_observation, variable_names, shap_values, timestamps, aggregation_method)
    return result, target_fun, baseline_fun, timestamps


def shap_kernel_explainer(explainer, new_observation, function_type, aggregation_method, timestamps, **kwargs):
    target_fun = explainer.predict(new_observation, function_type)[0]
    all_functions = explainer.predict(explainer.data, function_type)

    if timestamps is None:
        target_fun = target_fun.y
        all_functions_vals = [f.y for f in all_functions]
        timestamps = all_functions[0].x
    else:
        target_fun = target_fun(timestamps)
        all_functions_vals = [f(timestamps) for f in all_functions]
    baseline_fun = np.mean(all_functions_vals, axis=0)

    def predict_function(X):
        all_functions = explainer.predict(X, function_type)
        preds = np.array([f(timestamps) for f in all_functions])
        return preds

    # as shap convert pd.DataFrame to np.array
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        exp = shap.KernelExplainer(predict_function, explainer.data, **kwargs)
        res = exp.shap_values(new_observation)

    shap_values = np.vstack(res).T

    variable_names = explainer.data.columns
    result = prepare_result_df(new_observation, variable_names, shap_values, timestamps, aggregation_method)
    return result, target_fun, baseline_fun, timestamps


def shap_kernel(
    explainer, new_observation, function_type, aggregation_method, timestamps, max_shap_value_inputs=np.inf
):
    p = new_observation.shape[1]

    # only one new_observation allowed
    target_fun = explainer.predict(new_observation, function_type)[0]
    all_functions = explainer.predict(explainer.data, function_type)

    if timestamps is None:
        target_fun = target_fun.y
        all_functions_vals = [f.y for f in all_functions]
        timestamps = all_functions[0].x
    else:
        target_fun = target_fun(timestamps)
        all_functions_vals = [f(timestamps) for f in all_functions]
    baseline_f = np.mean(all_functions_vals, axis=0)

    if 2**p < max_shap_value_inputs:
        simplified_inputs = [list(z) for z in itertools.product(range(2), repeat=p)]
    else:
        simplified_inputs = list(np.random.randint(0, 2, size=(max_shap_value_inputs, p)))
        print(
            f"Approximate Survival Shapley will sample only {max_shap_value_inputs} values instead of 2**{p} for Exact Shapley"
        )

    kernel_weights = generate_shap_kernel_weights(simplified_inputs, p)
    shap_values, r2 = calculate_shap_values(
        explainer,
        function_type,
        baseline_f,
        explainer.data,
        simplified_inputs,
        kernel_weights,
        new_observation,
        timestamps,
    )

    variable_names = explainer.data.columns
    result = prepare_result_df(new_observation, variable_names, shap_values, timestamps, aggregation_method)
    return result, target_fun, baseline_f, timestamps, r2


def generate_shap_kernel_weights(simplified_inputs, num_variables):
    weights = []
    for coalition_vector in simplified_inputs:
        num_available_variables = np.count_nonzero(coalition_vector)
        if num_available_variables == 0 or num_available_variables == num_variables:
            weights.append(1e9)
        else:
            weights.append(
                (num_variables - 1)
                / (
                    math.comb(num_variables, num_available_variables)
                    * num_available_variables
                    * (num_variables - num_available_variables)
                )
            )
    return weights


def make_prediction_for_simplified_input(model, function_type, data, simplified_inputs, new_observation, timestamps):
    preds = np.zeros((len(simplified_inputs), len(timestamps)))
    for i, mask in enumerate(simplified_inputs):
        X_tmp = pd.DataFrame(np.where(mask, new_observation, data), columns=data.columns)
        preds[i,] = calculate_mean_function(model, function_type, X_tmp, timestamps)
    return preds


def calculate_shap_values(
    model,
    function_type,
    avg_function,
    data,
    simplified_inputs,
    shap_kernel_weights,
    new_observation,
    timestamps,
):
    W = np.diag(shap_kernel_weights)
    X = np.array(simplified_inputs)
    R = np.linalg.inv(X.T @ W @ X) @ (X.T @ W)
    y = (
        make_prediction_for_simplified_input(model, function_type, data, simplified_inputs, new_observation, timestamps)
        - avg_function
    )
    shap_values = R @ y
    y_pred = X @ shap_values
    r2 = [None] * y.shape[1]
    for i in range(y.shape[1]):
        r2[i] = r2_score(y[:, i], y_pred[:, i], sample_weight=shap_kernel_weights)
    return shap_values, r2


def shap_sampling(
    explainer,
    new_observation,
    function_type,
    path,
    B,
    random_state,
    aggregation_method,
    timestamps,
    exact=False,
):
    p = new_observation.shape[1]

    # only one new_observation allowed
    target_fun = explainer.predict(new_observation, function_type)[0]
    all_functions = explainer.predict(explainer.data, function_type)

    if timestamps is None:
        target_fun = target_fun.y
        all_functions_vals = [f.y for f in all_functions]
        timestamps = all_functions[0].x
    else:
        target_fun = target_fun(timestamps)
        all_functions_vals = [f(timestamps) for f in all_functions]
    baseline_f = np.mean(all_functions_vals, axis=0)

    if exact:
        permutations = [list(perm) for perm in itertools.permutations(np.arange(p), p)]
        B = len(permutations)
    else:
        permutations = [None] * B

    np.random.seed(random_state)
    result_list = [
        iterate_paths(
            explainer,
            function_type,
            explainer.data,
            new_observation,
            timestamps,
            p,
            b + 1,
            np.random,
            permutations[b],
        )
        for b in tqdm(range(B))
    ]

    result = pd.concat(result_list)

    if path is not None:
        if isinstance(path, str) and path == "average":
            average_changes = (
                result.groupby("variable_str").mean(numeric_only=True).iloc[:, 2:]
            )  # choose predictions, not variable value and B
            average_changes["aggregated_change"] = aggregate_change(average_changes, aggregation_method, timestamps)
            average_changes = average_changes.sort_values("aggregated_change", ascending=False)
            result_average = result_list[0].set_index("variable_str").reindex(average_changes.index).reset_index()
            result_average = result_average.assign(B=0)
            result_average.iloc[:, 4:] = average_changes.drop("aggregated_change", axis=1).values
            result_average.insert(4, "aggregated_change", average_changes["aggregated_change"].values)
            result = pd.concat((result_average, result), axis=0)
        else:
            tmp = get_single_random_path(
                explainer,
                function_type,
                explainer.data,
                new_observation,
                timestamps,
                path,
                0,
            )
            result = pd.concat((result, tmp))

    return result, target_fun, baseline_f, timestamps


def iterate_paths(model, function_type, data, new_observation, timestamps, p, b, rng, path):
    if path is None:
        random_path = rng.choice(np.arange(p), p, replace=False)
    else:
        random_path = path
    return get_single_random_path(model, function_type, data, new_observation, timestamps, random_path, b)


def get_single_random_path(model, function_type, data, new_observation, timestamps, random_path, b):
    current_data = deepcopy(data)
    yhats = [None] * (len(random_path) + 1)
    yhats[0] = calculate_mean_function(model, function_type, current_data, timestamps)
    for i, candidate in enumerate(random_path):
        current_data.iloc[:, candidate] = new_observation.iloc[0, candidate]
        yhats[i + 1] = calculate_mean_function(model, function_type, current_data, timestamps)

    diffs = np.diff(yhats, axis=0)

    variable_names = data.columns[random_path]

    new_observation_f = new_observation.loc[:, variable_names].apply(lambda x: nice_format(x.iloc[0]))

    result_diffs = pd.DataFrame(diffs.tolist(), columns=[" = ".join(["t", str(time)]) for time in timestamps])
    result_meta = pd.DataFrame(
        {
            "variable_str": [" = ".join(pair) for pair in zip(variable_names, new_observation_f)],
            "variable_name": variable_names,
            "variable_value": new_observation.loc[:, variable_names].values.reshape(
                -1,
            ),
            "B": b,
        }
    )

    return pd.concat([result_meta, result_diffs], axis=1)


def prepare_result_df(new_observation, variable_names, shap_values, timestamps, aggregation_method):
    new_observation_f = new_observation.apply(lambda x: nice_format(x.iloc[0]))
    result_shap = pd.DataFrame(shap_values, columns=[" = ".join(["t", str(time)]) for time in timestamps])
    result_meta = pd.DataFrame(
        {
            "variable_str": [" = ".join(pair) for pair in zip(variable_names, new_observation_f)],
            "variable_name": variable_names,
            "variable_value": new_observation.values.reshape(
                -1,
            ),
            "B": 0,
            "aggregated_change": aggregate_change(result_shap, aggregation_method, timestamps),
        }
    )
    return pd.concat([result_meta, result_shap], axis=1)


def aggregate_change(average_changes, aggregation_method, timestamps):
    if aggregation_method == "sum_of_squares":
        return np.sum(average_changes.values**2, axis=1)
    if aggregation_method == "max_abs":
        return np.max(np.abs(average_changes.values), axis=1)
    if aggregation_method == "mean_abs":
        return np.mean(np.abs(average_changes.values), axis=1)
    if aggregation_method == "integral":
        return trapezoid(np.abs(average_changes.values), timestamps)


def calculate_mean_function(model, function_type, data, timestamps):
    all_functions = model.predict(data, function_type)
    all_function_vals = [f(timestamps) for f in all_functions]
    return np.mean(all_function_vals, axis=0)


def nice_format(x):
    return str(x) if isinstance(x, (str, np.str_)) else str(float(signif(x)))


def signif(x, p=4):
    x = np.asarray(x)
    x_positive = np.where(np.isfinite(x) & (x != 0), np.abs(x), 10 ** (p - 1))
    mags = 10 ** (p - 1 - np.floor(np.log10(x_positive)))
    return np.round(x * mags) / mags


def calculate_risk_table(ticks, event_times, event_ind):
    n_at_risk = []
    n_censored = []
    n_events = []
    for i in ticks:
        n_at_risk.append((event_times > i).sum())
        n_events.append(event_ind[event_times <= i].sum())
        n_censored.append((~event_ind[event_times <= i]).sum())
    return n_at_risk, n_events, n_censored


def check_new_observation(new_observation, explainer):
    new_observation_ = deepcopy(new_observation)
    if isinstance(new_observation_, pd.Series):
        new_observation_ = new_observation_.to_frame().T
        new_observation_.columns = explainer.data.columns
    elif isinstance(new_observation_, np.ndarray):
        if new_observation_.ndim == 1:
            # make 1D array 2D
            new_observation_ = new_observation_.reshape((1, -1))
        elif new_observation_.ndim > 2:
            raise ValueError("Wrong new_observation dimension")
        elif new_observation.shape[0] != 1:
            raise ValueError("Wrong new_observation dimension")

        new_observation_ = pd.DataFrame(new_observation_)
        new_observation_.columns = explainer.data.columns

    elif isinstance(new_observation_, list):
        new_observation_ = pd.DataFrame(new_observation_).T
        new_observation_.columns = explainer.data.columns

    elif isinstance(new_observation_, pd.DataFrame):
        if new_observation.shape[0] != 1:
            raise ValueError("Wrong new_observation dimension")

        new_observation_.columns = explainer.data.columns
    else:
        raise TypeError("new_observation must be a numpy.ndarray or pandas.Series or pandas.DataFrame")

    if pd.api.types.is_bool_dtype(new_observation_.index):
        raise ValueError("new_observation index is of boolean type")

    return new_observation_



import numpy as np
import pandas as pd

from ..predict_explanations.utils import prepare_result_df
from ..predict_explanations.object import PredictSurvSHAP
from tqdm import tqdm
import matplotlib.pyplot as plt
from statsmodels.graphics.functional import fboxplot
from scipy.integrate import trapezoid
from sksurv.ensemble import RandomSurvivalForest
import shap
import warnings


def calculate_individual_explanations(
    explainer,
    new_observations,
    function_type,
    path,
    B,
    max_shap_value_inputs,
    random_state,
    calculation_method,
    aggregation_method,
    timestamps,
    save_individual_explanations,
    **kwargs,
):
    individual_explanations = []
    concatenated_results = pd.DataFrame()
    preds = explainer.predict(explainer.data, function_type)

    if timestamps is None:
        timestamps = preds[0].x
        preds = [f.y for f in preds]
    else:
        if calculation_method == "treeshap":
            if not isinstance(explainer.model, RandomSurvivalForest):
                raise TypeError("explained model must be of class sksurv.ensemble.RandomSurvivalForest")
            warnings.warn(
                "timestamps are ignored for calculation_method = 'treeshap' \n SurvSHAP(t) values are calculated for explainer.model.unique_times_"
            )
            timestamps = explainer.model.unique_times_
            preds = [f.y for f in preds]
        else:
            preds = [f(timestamps) for f in preds]

    baseline_f = np.mean(preds, axis=0)

    if calculation_method in ["shap_kernel", "treeshap"]:
        if calculation_method == "shap_kernel":

            def predict_function(X):
                all_functions = explainer.predict(X, function_type)
                preds = np.array([f(timestamps) for f in all_functions])
                return preds

            # as shap convert pd.DataFrame to np.array
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=UserWarning)
                exp = shap.KernelExplainer(predict_function, explainer.data, **kwargs)
                res = exp.shap_values(new_observations)
            tmp = np.dstack(res).flatten()

        elif calculation_method == "treeshap":
            if function_type == "sf":
                start_index = 1
            elif function_type == "chf":
                start_index = 0

            n_estimators = len(explainer.model.estimators_)
            tree_ensemble_model = {"trees": [estimator.tree_ for estimator in explainer.model.estimators_]}

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=UserWarning)
                exp = shap.TreeExplainer(tree_ensemble_model, explainer.data, **kwargs)
                res = exp.shap_values(new_observations, **kwargs)
            res = res[start_index::2]
            tmp = np.dstack(res).flatten() / n_estimators

        new_observations_shape = new_observations.shape
        tmp = tmp.reshape(new_observations_shape[0] * new_observations_shape[1], len(timestamps))
        variable_names = explainer.data.columns

        exp_y_names = explainer.y.dtype.names
        event_inds = explainer.y[exp_y_names[0]]
        event_times = explainer.y[exp_y_names[1]]

        for i in range(new_observations_shape[0]):
            survSHAP_obj = PredictSurvSHAP(
                function_type=function_type,
                calculation_method=calculation_method,
                aggregation_method=aggregation_method,
                random_state=random_state,
            )
            survSHAP_obj.result = prepare_result_df(
                new_observations.iloc[[i]],
                variable_names,
                tmp[(new_observations_shape[1] * i) : (new_observations_shape[1] * (i + 1)), :],
                timestamps,
                aggregation_method,
            )
            survSHAP_obj.simplified_result = survSHAP_obj.result[survSHAP_obj.result["B"] == 0].iloc[:, 1:5]
            survSHAP_obj.predicted_function = preds[i]
            survSHAP_obj.baseline_function = baseline_f
            survSHAP_obj.timestamps = timestamps

            survSHAP_obj.event_inds = event_inds
            survSHAP_obj.event_times = event_times

            if save_individual_explanations:
                individual_explanations.append(survSHAP_obj)
            tmp_results = survSHAP_obj.result
            tmp_results.insert(5, "index", i)
            concatenated_results = pd.concat((concatenated_results, tmp_results), axis=0)
    else:
        for i in tqdm(range(len(new_observations))):
            survSHAP_obj = PredictSurvSHAP(
                function_type=function_type,
                path=path,
                B=B,
                max_shap_value_inputs=max_shap_value_inputs,
                calculation_method=calculation_method,
                aggregation_method=aggregation_method,
                random_state=random_state,
            )
            survSHAP_obj.fit(explainer, new_observations.iloc[[i]], timestamps)
            if save_individual_explanations:
                individual_explanations.append(survSHAP_obj)
            tmp_results = survSHAP_obj.result
            tmp_results.insert(5, "index", i)
            concatenated_results = pd.concat((concatenated_results, tmp_results), axis=0)
    if timestamps is None:
        timestamps = survSHAP_obj.timestamps
    return concatenated_results, individual_explanations, timestamps


def create_boxplot_with_outliers(variable, full_result, wfactor=3):
    boxplot_data = full_result[(full_result["variable_name"] == variable) & (full_result["B"] == 0)].iloc[:, 6:].values
    fbxplt = fboxplot(boxplot_data, wfactor=wfactor)
    plt.close()
    outliers_ids = fbxplt[3]
    without_outliers_data = np.delete(boxplot_data, outliers_ids, axis=0)
    upper_whisker = np.max(without_outliers_data, axis=0)
    lower_whisker = np.min(without_outliers_data, axis=0)
    median_idx = fbxplt[2][0]
    return outliers_ids, median_idx, upper_whisker, lower_whisker


def aggregate_change(average_changes, aggregation_method, timestamps):
    if aggregation_method == "sum_of_squares":
        return np.sum(average_changes**2, axis=1)
    if aggregation_method == "max":
        return np.max(average_changes, axis=1)
    if aggregation_method == "mean":
        return np.mean(average_changes, axis=1)
    if aggregation_method == "integral":
        return trapezoid(average_changes.values, timestamps)


def calculate_risk_table(ticks, times, event_ind):
    n_at_risk = []
    n_censored = []
    n_events = []
    for i in ticks:
        n_at_risk.append((times > i).sum())
        n_events.append(event_ind[times <= i].sum())
        n_censored.append((~event_ind[times <= i]).sum())
    return n_at_risk, n_events, n_censored



import numpy as np
import pandas as pd
import matplotlib.ticker
import plotly.subplots
import plotly.graph_objects
from .utils import calculate_risk_table, create_boxplot_with_outliers
from matplotlib import cm
from matplotlib.colors import LinearSegmentedColormap, Normalize
import itertools


def tooltip_text(row):
    return f"<b>{row.variable_name}</b><br>" + f"Value: {row.variable_value:.6f}<br>"


def check_y_range(max_y, min_y, new_vector):
    tmp_max = np.max(new_vector)
    tmp_min = np.min(new_vector)
    if tmp_max > max_y:
        max_y = tmp_max
    if tmp_min < min_y:
        min_y = tmp_min
    return max_y, min_y


def insert_zeros_to_line(x_vals, y_vals):
    put_zero = np.where(np.diff(np.sign(y_vals)) != 0)[0]
    y1 = y_vals[put_zero]
    y2 = y_vals[put_zero + 1]

    x1 = x_vals[put_zero]
    x2 = x_vals[put_zero + 1]
    to_put_x = (x1 * abs(y2) + x2 * abs(y1)) / (abs(y1) + abs(y2))
    new_y = np.insert(y_vals, put_zero + 1, 0)
    new_x = np.insert(x_vals, put_zero + 1, to_put_x)

    return new_x, new_y


def model_plot_mean_abs_shap_values(
    result,
    timestamps,
    event_inds,
    event_times,
    variables,
    max_vars=10,
    x_range=None,
    kind="default",
    show_risk_table=True,
    title=None,
    show=True,
):
    df_prepared_to_plot = result.copy()
    if kind == "ratio":
        df_prepared_to_plot.iloc[:, 5:] = df_prepared_to_plot.iloc[:, 5:] / df_prepared_to_plot.iloc[:, 5:].abs().sum()
    if variables is not None:
        df_prepared_to_plot = df_prepared_to_plot[df_prepared_to_plot["variable_name"].isin(variables)]
    else:
        df_prepared_to_plot = (
            df_prepared_to_plot.sort_values("aggregated_change", ascending=False)
            .iloc[:max_vars]
            .reset_index(drop=True)
            .copy()
        )

    fig = plotly.subplots.make_subplots(rows=2, cols=1, print_grid=False, shared_xaxes=True)
    max_y, min_y = -1, 1
    fig.add_trace(
        plotly.graph_objs.Scatter(
            x=timestamps,
            y=[0] * len(timestamps),
            mode="lines",
            line_color="#4378bf",
            line_width=3,
            hovertemplate="<b>Average SF</b><br>" + "Time: %{x}<br>" + "<extra></extra>",
            hoverinfo="text",
        ),
        row=1,
        col=1,
    )

    keyword = "SHAP value: " if kind == "default" else "normalized SHAP value: "

    colors = ["#46bac2", "#ae2c87", "#ffa58c", "#8bdcbe", "#f05a71", "#FED61E", "#FED61E"]
    colors = itertools.cycle(colors)

    for index, row in df_prepared_to_plot.iterrows():
        y_vals = row[5:].values
        varval = row["variable_value"]
        fig.add_trace(
            plotly.graph_objs.Scatter(
                x=timestamps,
                y=y_vals,
                mode="lines",
                line_color=next(colors),
                line_width=2,
                hovertemplate="<b>"
                + str(row["variable_name"])
                + "</b><br>"
                + f"(Avg) variable value: {varval:.6f}"
                + "<br>"
                "Time: %{x}<br>" + keyword + "%{y:.6f}<extra></extra>",
                hoverinfo="text",
            ),
            row=1,
            col=1,
        )
        max_y, min_y = check_y_range(max_y, min_y, y_vals)

    min_y = min_y - 0.05 * abs(min_y)
    max_y = max_y + 0.05 * abs(max_y)

    if x_range is None:
        x_range = [min(timestamps) - max(timestamps) * 0.025, max(timestamps) * 1.025]

    ticker = matplotlib.ticker.MaxNLocator(nbins=10, min_n_ticks=4, integer=True, prune="upper")
    ticks = ticker.tick_values(x_range[0], x_range[1])

    if show_risk_table:
        n_at_risk, n_events, n_censored = calculate_risk_table(ticks, event_times, event_inds)
        fig.append_trace(
            plotly.graph_objs.Scatter(
                x=ticks,
                y=[0.8] * len(ticks),
                text=n_at_risk,
                mode="text",
                showlegend=False,
            ),
            2,
            1,
        )
        fig.append_trace(
            plotly.graph_objs.Scatter(
                x=ticks,
                y=[0.5] * len(ticks),
                text=n_events,
                mode="text",
                showlegend=False,
            ),
            2,
            1,
        )
        fig.append_trace(
            plotly.graph_objs.Scatter(
                x=ticks,
                y=[0.2] * len(ticks),
                text=n_censored,
                mode="text",
                showlegend=False,
            ),
            2,
            1,
        )

    fig.update_xaxes(
        {
            "matches": None,
            "showticklabels": True,
            "title": "time",
            "title_standoff": 0,
            "type": "linear",
            "gridwidth": 2,
            "zeroline": False,
            "automargin": True,
            "tickmode": "array",
            "tickvals": ticks,
            "ticktext": np.around(ticks, 2),
            "tickcolor": "white",
            "ticklen": 3,
            "fixedrange": True,
            "range": x_range,
        }
    ).update_yaxes(
        {
            "type": "linear",
            "gridwidth": 2,
            "zeroline": False,
            "automargin": True,
            "ticks": "outside",
            "tickcolor": "white",
            "ticklen": 3,
            "fixedrange": True,
            "range": [min_y, max_y],
        }
    ).update_layout(
        {"showlegend": False, "template": "none", "margin_pad": 6, "margin_l": 110}
    ).update_layout(
        yaxis2={
            "tickvals": [0.2, 0.5, 0.8],
            "ticktext": ["censored", "events", "at risk"],
            "showgrid": False,
            "domain": [0.0, 0.2],
            "range": [0, 1],
        },
        xaxis2={"visible": False},
    )

    if kind == "ratio":
        fig.update_layout(
            yaxis1={"domain": [0.35, 1], "title": "normalized mean(|SHAP value|)"},
        )
    else:
        fig.update_layout(
            yaxis1={"domain": [0.35, 1], "title": "mean(|SHAP value|)"},
        )

    fig.update_layout(
        title_text=title,
        title_x=0.15,
        font={"color": "#371ea3"},
        template="none",
        margin={"t": 78, "b": 71, "r": 30},
    )

    if show:
        fig.show(
            config={
                "displaylogo": False,
                "staticPlot": False,
                "toImageButtonOptions": {
                    "height": None,
                    "width": None,
                },
                "modeBarButtonsToRemove": [
                    "sendDataToCloud",
                    "lasso2d",
                    "autoScale2d",
                    "select2d",
                    "zoom2d",
                    "pan2d",
                    "zoomIn2d",
                    "zoomOut2d",
                    "resetScale2d",
                    "toggleSpikelines",
                    "hoverCompareCartesian",
                    "hoverClosestCartesian",
                ],
            }
        )

    else:
        return fig


def model_plot_shap_lines_for_all_individuals(
    full_result,
    timestamps,
    event_inds,
    event_times,
    variable,
    x_range=None,
    kind="default",
    boxplot=False,
    wfactor=3,
    show_risk_table=True,
    show=True,
    title=None,
):
    if variable not in full_result["variable_name"].unique():
        raise ValueError("Variable not in the result")
    df_prepared_to_plot = full_result[full_result["variable_name"] == variable]
    df_prepared_to_plot = df_prepared_to_plot[df_prepared_to_plot["B"] == 0]
    if kind == "ratio":
        df_prepared_to_plot.iloc[:, 6:] = df_prepared_to_plot.iloc[:, 6:] / df_prepared_to_plot.iloc[:, 6:].abs().sum()
    df_prepared_to_plot = df_prepared_to_plot.reset_index(drop=True)
    if ~boxplot:
        cmap = LinearSegmentedColormap.from_list(
            "dalex",
            [
                "#8bdcbe",
                "#77d1be",
                "#61c5c0",
                "#46bac2",
                "#45a6c4",
                "#4590c4",
                "#4378bf",
                "#415fb9",
                "#3d42af",
                "#371ea3",
            ],
            N=100,
        )
        minima = np.min(df_prepared_to_plot["variable_value"])
        maxima = np.max(df_prepared_to_plot["variable_value"])
        norm = matplotlib.colors.Normalize(vmin=minima, vmax=maxima, clip=True)
        mapper = cm.ScalarMappable(norm=norm, cmap=cmap)
        mapped = mapper.to_rgba(df_prepared_to_plot["variable_value"])
        mapped = np.apply_along_axis(lambda x: matplotlib.colors.rgb2hex(x[:3]), 1, mapped)
        df_prepared_to_plot["index"] = mapped

    fig = plotly.subplots.make_subplots(rows=2, cols=1, print_grid=False, shared_xaxes=True)

    max_y, min_y = -1, 1

    keyword = "SHAP value: " if kind == "default" else "normalized SHAP value: "

    if boxplot:
        (
            outliers_ids,
            median_idx,
            upper_whisker,
            lower_whisker,
        ) = create_boxplot_with_outliers(variable, full_result, wfactor)
    for i, row in df_prepared_to_plot.iterrows():
        x_vals = timestamps
        y_vals = row[6:].values
        new_x, new_y = insert_zeros_to_line(x_vals, y_vals)
        if boxplot:
            if i in outliers_ids:
                line_color = "#ae2c87"
                line_width = 1.5
            elif i == median_idx:
                line_color = "#ffa58c"
                line_width = 2
            else:
                line_color = "lightgrey"
                line_width = 0.5
        else:
            line_color = row["index"]
            line_width = 0.5
        fig.add_trace(
            plotly.graph_objs.Scatter(
                x=new_x,
                y=new_y,
                mode="lines",
                line_color=line_color,
                line_width=line_width,
                hovertemplate="<b>"
                + str(row["variable_name"])
                + "</b><br>"
                + "Variable value: "
                + str(row["variable_value"])
                + "<br>"
                "Time: %{x}<br>" + keyword + "%{y:.6f}<extra></extra>",
                hoverinfo="text",
            ),
            row=1,
            col=1,
        )
        max_y, min_y = check_y_range(max_y, min_y, y_vals)
    if boxplot:
        fig.add_trace(
            plotly.graph_objs.Scatter(
                x=timestamps,
                y=lower_whisker,
                mode="lines",
                line_color="#ffa58c",
                line_width=2,
            ),
            row=1,
            col=1,
        )
        fig.add_trace(
            plotly.graph_objs.Scatter(
                x=timestamps,
                y=upper_whisker,
                mode="lines",
                line_color="#ffa58c",
                line_width=2,
            ),
            row=1,
            col=1,
        )
    min_y = min_y - 0.05 * abs(min_y)
    max_y = max_y + 0.05 * abs(max_y)

    if x_range is None:
        x_range = [min(timestamps) - max(timestamps) * 0.025, max(timestamps) * 1.025]

    ticker = matplotlib.ticker.MaxNLocator(nbins=10, min_n_ticks=4, integer=True, prune="upper")
    ticks = ticker.tick_values(x_range[0], x_range[1])

    if show_risk_table:
        n_at_risk, n_events, n_censored = calculate_risk_table(ticks, event_times, event_inds)
        fig.append_trace(
            plotly.graph_objs.Scatter(
                x=ticks,
                y=[0.8] * len(ticks),
                text=n_at_risk,
                mode="text",
                showlegend=False,
            ),
            2,
            1,
        )
        fig.append_trace(
            plotly.graph_objs.Scatter(
                x=ticks,
                y=[0.5] * len(ticks),
                text=n_events,
                mode="text",
                showlegend=False,
            ),
            2,
            1,
        )
        fig.append_trace(
            plotly.graph_objs.Scatter(
                x=ticks,
                y=[0.2] * len(ticks),
                text=n_censored,
                mode="text",
                showlegend=False,
            ),
            2,
            1,
        )

    fig.update_xaxes(
        {
            "matches": None,
            "showticklabels": True,
            "title": "time",
            "title_standoff": 0,
            "type": "linear",
            "gridwidth": 2,
            "zeroline": False,
            "automargin": True,
            "tickmode": "array",
            "tickvals": ticks,
            "ticktext": np.around(ticks, 2),
            "tickcolor": "white",
            "ticklen": 3,
            "fixedrange": True,
            "range": x_range,
        }
    ).update_yaxes(
        {
            "type": "linear",
            "gridwidth": 2,
            "zeroline": False,
            "automargin": True,
            "ticks": "outside",
            "tickcolor": "white",
            "ticklen": 3,
            "fixedrange": True,
            "range": [min_y, max_y],
        }
    ).update_layout(
        {"showlegend": False, "template": "none", "margin_pad": 6, "margin_l": 110}
    ).update_layout(
        yaxis2={
            "tickvals": [0.2, 0.5, 0.8],
            "ticktext": ["censored", "events", "at risk"],
            "showgrid": False,
            "domain": [0.0, 0.2],
            "range": [0, 1],
        },
        xaxis2={"visible": False},
    )
    if kind == "ratio":
        fig.update_layout(
            yaxis1={"domain": [0.35, 1], "title": "normalized SHAP value"},
        )
    else:
        fig.update_layout(
            yaxis1={"domain": [0.35, 1], "title": "SHAP value"},
        )

    fig.update_layout(
        title_text=title,
        title_x=0.15,
        font={"color": "#371ea3"},
        template="none",
        margin={"t": 78, "b": 71, "r": 30},
    )

    if show:
        fig.show(
            config={
                "displaylogo": False,
                "staticPlot": False,
                "toImageButtonOptions": {
                    "height": None,
                    "width": None,
                },
                "modeBarButtonsToRemove": [
                    "sendDataToCloud",
                    "lasso2d",
                    "autoScale2d",
                    "select2d",
                    "zoom2d",
                    "pan2d",
                    "zoomIn2d",
                    "zoomOut2d",
                    "resetScale2d",
                    "toggleSpikelines",
                    "hoverCompareCartesian",
                    "hoverClosestCartesian",
                ],
            }
        )

    else:
        return fig



import numpy as np
import matplotlib.ticker
import plotly.subplots
import plotly.graph_objects
import itertools
from .utils import calculate_risk_table


def tooltip_text(row):
    return f"<b>{row.variable_name}</b><br>" + f"Value: {float(row.variable_value):.6g}<br>"


def check_y_range(max_y, min_y, new_vector):
    tmp_max = np.max(new_vector)
    tmp_min = np.min(new_vector)
    if tmp_max > max_y:
        max_y = tmp_max
    if tmp_min < min_y:
        min_y = tmp_min
    return max_y, min_y


def insert_zeros_to_line(x_vals, y_vals, baseline_function):
    put_zero = np.where(np.diff(np.sign(y_vals)) != 0)[0]
    y1 = y_vals[put_zero]
    y2 = y_vals[put_zero + 1]

    x1 = x_vals[put_zero]
    x2 = x_vals[put_zero + 1]
    to_put_x = (x1 * abs(y2) + x2 * abs(y1)) / (abs(y1) + abs(y2))
    bsf1 = baseline_function[put_zero]
    bsf2 = baseline_function[put_zero + 1]
    to_put_baseline_function = (bsf1 * abs(y2) + bsf2 * abs(y1)) / (abs(y1) + abs(y2))
    new_x = np.insert(x_vals, put_zero + 1, to_put_x)
    new_y = np.insert(y_vals, put_zero + 1, 0)
    new_baseline_function = np.insert(baseline_function, put_zero + 1, to_put_baseline_function)

    return new_x, new_y, new_baseline_function


def predict_plot(
    result,
    predicted_function,
    baseline_function,
    timestamps,
    event_inds,
    event_times,
    y_true_ind=None,
    y_true_time=None,
    max_vars=10,
    variables=None,
    x_range=None,
    kind="default",
    add_to_baseline=False,
    show_prediction=True,
    show_overall_change=False,
    show_y_info=True,
    show_risk_table=True,
    title=None,
    show=True,
):
    # get averaged shap values
    final_result = result[result["B"] == 0]
    if kind == "ratio":
        final_result.iloc[:, 5:] = np.nan_to_num(
            final_result[final_result["B"] == 0].iloc[:, 5:]
            / final_result[final_result["B"] == 0].iloc[:, 5:].abs().sum()
        )
        add_to_baseline = False
        show_prediction = False
        show_overall_change = False

    # choose variables
    if variables is not None:
        df_prepared_to_plot = final_result[final_result["variable_name"].isin(variables)].copy()
    else:
        variables = (
            final_result.sort_values("aggregated_change", key=abs, ascending=False).iloc[:max_vars].variable_name
        )
        df_prepared_to_plot = final_result[final_result["variable_name"].isin(variables)].copy()

    # add tooltips
    df_prepared_to_plot["tooltip_text"] = df_prepared_to_plot.apply(lambda row: tooltip_text(row), axis=1)

    fig = plotly.subplots.make_subplots(rows=2, cols=1, print_grid=False, shared_xaxes=True)

    max_y, min_y = -1, 1

    # base/background function
    if add_to_baseline:
        # average sf
        fig.add_trace(
            plotly.graph_objs.Scatter(
                x=timestamps,
                y=baseline_function,
                mode="lines",
                line_color="#4378bf",
                line_width=3,
                hovertemplate="<b>Average SF</b><br>" + "Time: %{x}<br>" + "Avg SF value: %{y:.6f}<extra></extra>",
                hoverinfo="text",
            ),
            row=1,
            col=1,
        )
        max_y, min_y = check_y_range(max_y, min_y, baseline_function)

    else:
        # average sf as 0
        fig.add_trace(
            plotly.graph_objs.Scatter(
                x=timestamps,
                y=[0] * len(timestamps),
                mode="lines",
                line_color="#4378bf",
                line_width=3,
                text=baseline_function,
                hovertemplate="<b>Average SF</b><br>" + "Time: %{x}<br>" + "Avg SF value: %{text:.6f}<extra></extra>",
                hoverinfo="text",
            ),
            row=1,
            col=1,
        )

    start_idx_preds = np.where(df_prepared_to_plot.columns == f"t = {timestamps[0]}")[0][0]
    keyword = "SF value: " if add_to_baseline else "SHAP value: " if kind == "default" else "normalized SHAP value: "

    # SurvSHAP(t) curves for different variables

    # prepare colors
    colors = ["#46bac2", "#ae2c87", "#ffa58c", "#8bdcbe", "#f05a71", "#FED61E", "#FED61E"]
    colors = itertools.cycle(colors)

    for index, row in df_prepared_to_plot.iterrows():
        x_vals = timestamps
        y_vals = row[start_idx_preds:-1].values

        fig.add_trace(
            plotly.graph_objs.Scatter(
                x=x_vals,
                y=y_vals + baseline_function if add_to_baseline else y_vals,
                mode="lines",
                line_color=next(colors),
                hovertemplate=row["tooltip_text"] + "Time: %{x}<br>" + keyword + "%{y:.6f}<extra></extra>",
                hoverinfo="text",
            ),
            row=1,
            col=1,
        )
        new_vector = (
            row[start_idx_preds:-1].values + baseline_function if add_to_baseline else row[start_idx_preds:-1].values
        )
        max_y, min_y = check_y_range(max_y, min_y, new_vector)

    if add_to_baseline and show_prediction:
        fig.add_trace(
            plotly.graph_objs.Scatter(
                x=timestamps,
                y=predicted_function,
                mode="lines",
                line_color="#371ea3",
                line_width=3,
                hovertemplate="<b>Predicted SF</b><br>"
                + "Time: %{x}<br>"
                + "Predicted SF value: %{y:.6f}<extra></extra>",
                hoverinfo="text",
            ),
            row=1,
            col=1,
        )
        max_y, min_y = check_y_range(max_y, min_y, predicted_function)
    elif ~(add_to_baseline) and show_overall_change:
        fig.add_trace(
            plotly.graph_objs.Scatter(
                x=timestamps,
                y=predicted_function - baseline_function,
                mode="lines",
                line_color="#371ea3",
                line_width=3,
                hovertemplate="<b>SF change</b><br>" + "Time: %{x}<br>" + "SF change: %{y:.6f}<extra></extra>",
                hoverinfo="text",
            ),
            row=1,
            col=1,
        )
        max_y, min_y = check_y_range(max_y, min_y, predicted_function - baseline_function)

    min_y = min_y - 0.05 * abs(min_y)
    max_y = max_y + 0.05 * abs(max_y)

    if show_y_info and y_true_time is not None:
        status = "dead" if y_true_ind else "censored"
        fig.add_trace(
            plotly.graph_objs.Scatter(
                x=[y_true_time] * 100,
                y=np.linspace(min_y, max_y, num=100),
                mode="lines",
                line_color="#ae2c87",
                line_width=2,
                hovertemplate="<b>Outcome</b><br>"
                + "Time to event: %{x}<br>"
                + "Status: "
                + status
                + "<extra></extra>",
                hoverinfo="text",
            ),
            row=1,
            col=1,
        )

    if x_range is None:
        x_range = [min(timestamps) - max(timestamps) * 0.025, max(timestamps) * 1.025]

    ticker = matplotlib.ticker.MaxNLocator(nbins=10, min_n_ticks=4, integer=True, prune="upper")
    ticks = ticker.tick_values(x_range[0], x_range[1])
    if show_risk_table:
        n_at_risk, n_events, n_censored = calculate_risk_table(ticks, event_times, event_inds)
        fig.append_trace(
            plotly.graph_objs.Scatter(
                x=ticks,
                y=[0.8] * len(ticks),
                text=n_at_risk,
                mode="text",
                showlegend=False,
            ),
            2,
            1,
        )
        fig.append_trace(
            plotly.graph_objs.Scatter(
                x=ticks,
                y=[0.5] * len(ticks),
                text=n_events,
                mode="text",
                showlegend=False,
            ),
            2,
            1,
        )
        fig.append_trace(
            plotly.graph_objs.Scatter(
                x=ticks,
                y=[0.2] * len(ticks),
                text=n_censored,
                mode="text",
                showlegend=False,
            ),
            2,
            1,
        )

    fig.update_xaxes(
        {
            "matches": None,
            "showticklabels": True,
            "title": "time",
            "title_standoff": 0,
            "type": "linear",
            "gridwidth": 2,
            "zeroline": False,
            "automargin": True,
            "tickmode": "array",
            "tickvals": ticks,
            "ticktext": np.around(ticks, 2),
            "tickcolor": "white",
            "ticklen": 3,
            "fixedrange": True,
            "range": x_range,
        }
    ).update_yaxes(
        {
            "type": "linear",
            "gridwidth": 2,
            "zeroline": False,
            "automargin": True,
            "ticks": "outside",
            "tickcolor": "white",
            "ticklen": 3,
            "fixedrange": True,
            "range": [min_y, max_y],
        }
    ).update_layout(
        {"showlegend": False, "template": "none", "margin_pad": 6, "margin_l": 110}
    ).update_layout(
        yaxis2={
            "tickvals": [0.2, 0.5, 0.8],
            "ticktext": ["censored", "events", "at risk"],
            "showgrid": False,
            "domain": [0.0, 0.2],
            "range": [0, 1],
        },
        yaxis1={"domain": [0.35, 1]},
        xaxis2={"visible": False},
    )
    if add_to_baseline:
        fig.update_layout(yaxis1={"title": "survival function"})
    elif kind == "ratio":
        fig.update_layout(yaxis1={"title": "normalized SHAP value"})
    else:
        fig.update_layout(yaxis1={"title": "SHAP value"})

    fig.update_layout(
        title_text=title,
        title_x=0.15,
        font={"color": "#371ea3"},
        template="none",
        margin={"t": 78, "b": 71, "r": 30},
    )

    if show:
        fig.show(
            config={
                "displaylogo": False,
                "staticPlot": False,
                "toImageButtonOptions": {
                    "height": None,
                    "width": None,
                },
                "modeBarButtonsToRemove": [
                    "sendDataToCloud",
                    "lasso2d",
                    "autoScale2d",
                    "select2d",
                    "zoom2d",
                    "pan2d",
                    "zoomIn2d",
                    "zoomOut2d",
                    "resetScale2d",
                    "toggleSpikelines",
                    "hoverCompareCartesian",
                    "hoverClosestCartesian",
                ],
            }
        )

    else:
        return fig



from .predict_explanations.object import PredictSurvSHAP
from .model_explanations.object import ModelSurvSHAP
from .explainer import SurvivalModelExplainer

__version__ = "0.4.2"

__all__ = ["PredictSurvSHAP", "ModelSurvSHAP", "SurvivalModelExplainer"]



from .plot import predict_plot
import pandas as pd
import numpy as np
import pandas as pd
from .utils import check_new_observation, shap_kernel, shap_sampling, shap_kernel_explainer, shap_tree_explainer


class PredictSurvSHAP:
    def __init__(
        self,
        function_type="sf",
        calculation_method="kernel",
        aggregation_method="integral",
        path="average",
        B=25,
        max_shap_value_inputs=np.inf,
        random_state=None,
        exact=False,
    ):
        """Constructor for class PredictSurvSHAP

        Args:
            function_type (str, optional): Either "sf" representing survival function or "chf" representing cumulative hazard function. Type of function to be evaluated for explanation. Defaults to "sf".
            calculation_method (str, optional): Chooses the method for SurvSHAP(t) calculation. "shap_kernel" for shap.KernelExplainer, "kernel" for exact KernelSHAP, "sampling" for sampling method, or "treeshap" for shap.TreeExplainer. Defaults to "kernel".
            aggregation_method (str, optional): One of "sum_of_squares", "max_abs", "mean_abs" or "integral". Type of method  Defaults to "integral".
            path (list of int or str, optional): If specified, then attributions for this path will be plotted. Defaults to "average".
            B (int, optional): Number of random paths to calculate variable attributions. Defaults to 25.
            max_shap_value_inputs (int, optional): Maximum number of simplified inputs to be used for SurvSHAP(t) calculation. Defaults to np.inf (no limit). Lower values can be used to speed up calculation.
            random_state (int, optional): Set seed for random number generator. Defaults to None.
            exact (bool, optional): Calculates all paths. If this is set to True parameter B is overriden. Defaults to False.
        """
        self.function = function_type
        self.calculation_method = calculation_method
        self.aggregation_method = aggregation_method
        self.path = path
        self.B = B
        self.result = pd.DataFrame()
        self.timestamps = None
        self.predicted_function = None
        self.baseline_function = None
        self.changes = None
        self.random_state = random_state
        self.exact = exact
        self.event_inds = None  # full y
        self.event_times = None  # full y
        self.y_true_time = None  # for this instance
        self.y_true_ind = None  # for this instance
        self.r2 = None
        self.max_shap_value_inputs = max_shap_value_inputs

    def _repr_html_(self):
        return self.simplified_result._repr_html_()

    def fit(self, explainer, new_observation, timestamps=None, y_true=None):
        """Calculate SurvSHAP(t) for an observation

        Args:
            explainer (SurvivalModelExplainer): A wrapper object for the model to be explained.
            new_observation (pandas.DataFrame): A DataFrame with a single row, containing the observation to be explained.
            timestamps (numpy.Array, optional): An array of timestamps at which SurvSHAP(t) values should be calculated. Defaults to None.
            y_true (pandas.DataFrame, optional): A DataFrame containing the observed time and status of the explained observation. Used for plotting. Defaults to None.

        Raises:
            ValueError: if calculation_method is invalid
        """
        new_observation = check_new_observation(new_observation, explainer)
        names = explainer.y.dtype.names
        self.event_inds = explainer.y[names[0]]
        self.event_times = explainer.y[names[1]]
        if y_true is not None:
            self.y_true_ind = y_true[names[0]]
            self.y_true_time = y_true[names[1]]

        if self.calculation_method == "kernel":
            (
                self.result,
                self.predicted_function,
                self.baseline_function,
                self.timestamps,
                self.r2,
            ) = shap_kernel(
                explainer,
                new_observation,
                self.function,
                self.aggregation_method,
                timestamps,
                self.max_shap_value_inputs,
            )
        elif self.calculation_method == "sampling":
            (
                self.result,
                self.predicted_function,
                self.baseline_function,
                self.timestamps,
            ) = shap_sampling(
                explainer,
                new_observation,
                self.function,
                self.path,
                self.B,
                self.random_state,
                self.aggregation_method,
                timestamps,
                self.exact,
            )
        elif self.calculation_method == "shap_kernel":
            (
                self.result,
                self.predicted_function,
                self.baseline_function,
                self.timestamps,
            ) = shap_kernel_explainer(
                explainer,
                new_observation,
                self.function,
                self.aggregation_method,
                timestamps,
            )
        elif self.calculation_method == "treeshap":
            (
                self.result,
                self.predicted_function,
                self.baseline_function,
                self.timestamps,
            ) = shap_tree_explainer(
                explainer,
                new_observation,
                self.function,
                self.aggregation_method,
                timestamps,
                treeshap=True,
            )
        else:
            raise ValueError("calculation_method should be 'kernel', 'sampling', 'shap_kernel', or 'treeshap'")

        self.simplified_result = self.result[self.result["B"] == 0].iloc[:, 1:5]

    def plot(
        self,
        max_vars=10,
        variables=None,
        x_range=None,
        kind="default",
        add_to_baseline=False,
        show_prediction=True,
        show_overall_change=False,
        show_y_info=True,
        show_risk_table=True,
        title=None,
        show=True,
    ):
        """Plot SurvSHAP(t) explanation

        Args:
            max_vars (int, optional): Maximum number of most important variables to show on the plot. Defaults to 10.
            variables (array_like, optional): Names of variables to be included to be shown on the plot. If None, all variables are considered (and max_vars most important ones are shown). Defaults to None.
            x_range (array_like, optional): Array containing two numbers, start and end time to be plotted. Defaults to None.
            kind (str, optional): Either "default" or "ratio", if "ratio" then SurvSHAP(t) values are normalized by the sum of SurvSHAP(t) absolute values for all variables. Defaults to "default".
            add_to_baseline (bool, optional): If True the SurvSHAP(t) values are added to survival function, not plotted directly. Defaults to False.
            show_prediction (bool, optional): If True the black-box model's survival function prediction is plotted. Defaults to True.
            show_overall_change (bool, optional): If True, the overall change, compared to average is shown on the plot (black-box prediction - average prediction). Defaults to False.
            show_y_info (bool, optional): If True, the true observed time and status are included on the plot. Defaults to True.
            show_risk_table (bool, optional): If True, a table is shown below the x-axis showing how many observations observed the event, are at risk and how many are censored up to this time. Defaults to True.
            title (str, optional): A title of the plot. Defaults to None.
            show (bool, optional): If True, the plot is shown, if False a plotly Figure object is returned. Defaults to True.

        Returns:
            plotly.graph_objects.Figure: If show is `False` then a plotly Figure is returned.
        """

        return predict_plot(
            self.result,
            self.predicted_function,
            self.baseline_function,
            self.timestamps,
            self.event_inds,
            self.event_times,
            self.y_true_ind,
            self.y_true_time,
            max_vars,
            variables,
            x_range,
            kind,
            add_to_baseline,
            show_prediction,
            show_overall_change,
            show_y_info,
            show_risk_table,
            title,
            show,
        )



from .plot import model_plot_mean_abs_shap_values, model_plot_shap_lines_for_all_individuals
import pandas as pd
import numpy as np
import pandas as pd
from .utils import aggregate_change, calculate_individual_explanations
import warnings


class ModelSurvSHAP:
    def __init__(
        self,
        function_type="sf",
        calculation_method="kernel",
        aggregation_method="integral",
        path="average",
        B=25,
        max_shap_value_inputs=np.inf,
        random_state=None,
    ):
        """Constructor for class ModelSurvSHAP

        Args:
            function_type (str, optional): Either "sf" representing survival function or "chf" representing cumulative hazard function. Type of function to be evaluated for explanation. Defaults to "sf".
            calculation_method (str, optional): Chooses the method for SurvSHAP(t) calculation. "shap_kernel" for shap.KernelExplainer, "kernel" for exact KernelSHAP, "sampling" for sampling method, or "treeshap" for shap.TreeExplainer. Defaults to "kernel".
            aggregation_method (str, optional): One of "sum_of_squares", "max_abs", "mean_abs" or "integral". Type of method  Defaults to "integral".
            path (list of int or str, optional): If specified, then attributions for this path will be plotted. Defaults to "average".
            B (int, optional): Number of random paths to calculate variable attributions. Defaults to 25.
            max_shap_value_inputs (int, optional): Maximum number of simplified inputs to be used for SurvSHAP(t) calculation. Defaults to np.inf (no limit). Lower values can be used to speed up calculation.
            random_state (int, optional): Set seed for random number generator. Defaults to None.
        """
        self.explainer = None
        self.function_type = function_type
        self.calculation_method = calculation_method
        self.aggregation_method = aggregation_method
        self.path = path
        self.B = B
        self.max_shap_value_inputs = max_shap_value_inputs
        self.result = pd.DataFrame()
        self.timestamps = None
        self.random_state = random_state
        self.event_ind = None  # full y
        self.event_times = None  # full y
        self.full_result = None

    def _repr_html_(self):
        return self.result[self.result["B"] == 0]._repr_html_()

    def fit(self, explainer, new_observations=None, timestamps=None, save_individual_explanations=True, **kwargs):
        """Calculate SurvSHAP(t) for many new observations and aggregate results.

        Args:
            explainer (SurvivalModelExplainer): A wrapper object for the model to be explained.
            new_observations (pandas.DataFrame, optional): A DataFrame containing the observations to be explained. If None observations from explainer are explained. Defaults to None.
            timestamps (numpy.Array, optional): An array of timestamps at which SurvSHAP(t) values should be calculated. Defaults to None.
            save_individual_explanations (bool, optional): Whether to save PredictSurvSHAP objects (explanations for individual observations). Defaults to True.
            **kwargs (optional): Additional parameters passed for shap.KernelExplainer.
        """
        # based on original shap warning
        data_len = len(explainer.data)

        if data_len > 100 and self.calculation_method != "treeshap":
            warnings.warn(
                "Using "
                + str(data_len)
                + " background data samples could cause slower run times.\n"
                + "Consider using a smaller sample."
            )

        if new_observations is None:
            new_observations = explainer.data

        (
            self.full_result,
            self.individual_explanations,
            self.timestamps,
        ) = calculate_individual_explanations(
            explainer,
            new_observations,
            self.function_type,
            self.path,
            self.B,
            self.max_shap_value_inputs,
            self.random_state,
            self.calculation_method,
            self.aggregation_method,
            timestamps,
            save_individual_explanations,
            **kwargs
        )

        names = explainer.y.dtype.names
        self.event_ind = explainer.y[names[0]]
        self.event_times = explainer.y[names[1]]

        result = self.full_result.copy()

        result = (
            result[result["B"] == 0]
            .drop("variable_str", axis=1)
            .groupby(["variable_name"])
            .agg(lambda x: x.abs().mean())
            .reset_index()
        )

        result["aggregated_change"] = aggregate_change(result.iloc[:, 5:], self.aggregation_method, self.timestamps)
        result = result.sort_values("aggregated_change", ascending=False)
        self.result = result

    def get_mean_abs_shap_values(
        self,
        aggregation_method="sum_of_squares",
    ):
        result = self.full_result.copy()

        result = (
            result[result["B"] == 0]
            .drop("variable_str", axis=1)
            .groupby(["variable_name"])
            .agg(lambda x: x.abs().mean())
            .reset_index()
        )

        result["aggregated_change"] = aggregate_change(result.iloc[:, 5:], aggregation_method, self.timestamps)
        result = result.sort_values("aggregated_change", ascending=False)
        self.result = result

    def plot_mean_abs_shap_values(
        self,
        variables=None,
        max_vars=10,
        x_range=None,
        kind="default",
        show_risk_table=True,
        title=None,
        show=True,
    ):
        if len(self.result) == 0:
            self.get_mean_abs_shap_values()
        return model_plot_mean_abs_shap_values(
            self.result,
            self.timestamps,
            self.event_ind,
            self.event_times,
            variables,
            max_vars,
            x_range,
            kind,
            show_risk_table,
            title,
            show,
        )

    def plot_shap_lines_for_all_individuals(
        self,
        variable,
        x_range=None,
        kind="default",
        boxplot=False,
        wfactor=3,
        show_risk_table=True,
        show=True,
        title=None,
    ):
        return model_plot_shap_lines_for_all_individuals(
            self.full_result,
            self.timestamps,
            self.event_ind,
            self.event_times,
            variable,
            x_range,
            kind,
            boxplot,
            wfactor,
            show_risk_table,
            show,
            title,
        )
